/* Name: main.c
 * Author: Michael Ciuffo
 * Published under Creative Commons Attribution-NonCommercial 3.0 Unported license
 */

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>

uint8_t numbersB[10] PROGMEM = {0x00,0x20,0x00,0x00,0x20,0x00,0x00,0x20,0x00,0x00};
uint8_t numbersC[10] PROGMEM = {0x10,0x32,0x21,0x22,0x02,0x02,0x00,0x32,0x00,0x02};
uint8_t numbersD[10] PROGMEM = {0x00,0x01,0x00,0x00,0x01,0x02,0x02,0x00,0x00,0x00};
volatile uint8_t buffBTens = 0x00;
volatile uint8_t buffCTens = 0x00;
volatile uint8_t buffDTens = 0x00;
volatile uint8_t buffBOnes = 0x00;
volatile uint8_t buffCOnes = 0x00;
volatile uint8_t buffDOnes = 0x00;


volatile uint8_t displayOn = 0x01;
volatile uint16_t beepCount = 0x00;
volatile uint8_t threshold = 50;
volatile uint8_t interfaceTimeout=0;
volatile uint8_t sleepTime=0;
volatile uint8_t measureTime = 0;


//when ANDed with the current buffers, these will preserve the non-display related bytes
const uint8_t bMask = ~0x20;
const uint8_t cMask = ~0x33;
const uint8_t dMask = ~0x03;
const uint8_t cdigitEnableMask = ~0x0C;
uint8_t number = 0;

//Button masks to be used with Port D
const uint8_t buttonOneMask = 0x04;
const uint8_t buttonTwoMask = 0x08;

//Some timing constants to play with
const uint8_t maxInterfaceTimeout = 10; //time before device starts showing current measurement
const uint8_t debounceTime = 10;
const uint8_t maxSleepTime = 4; //time before device falls asleep after user interfaces in 8 second intervals
const uint8_t maxMeasureTime =4; //time between measurements in 8 second intervals


uint8_t takeMeasurement()
{
    PORTB = PORTB | 0X10; //Enable 3.0V supply
    _delay_ms(10);
    
    //configure ADC
    ADMUX = 0x27; //left adjust, AREF, ADC7
    ADCSRA = 0x87; //enable ADC, divide clock by 128

    ADCSRA|=0b01000000; //START CONVERSION
    while((ADCSRA & (0b01000000)))//WAIT FOR IT TO FINISH
        ADCSRA|=(0b00010000);//CLEAR FLAG
    PORTB = PORTB & ~0X10; //Disable 3.0V supply
    uint8_t temp = ADCH;
    
    ADCSRA = 0x00; //disable ADC

    
    return (temp>>1);
    
}

void incThreshold()
{
	if (threshold < 99)
		threshold++;
	return;
}
void decThreshold()
{
	if (threshold>0)
		threshold--;
	return;
}

void goTheFuckToSleep()
{
	//set all inputs to not draw any power
	uint8_t cDirection = DDRC;
	uint8_t bDirection = DDRB;
	uint8_t dDirection = DDRD;
	uint8_t cSet = PORTC;
	uint8_t dSet = PORTD;
	uint8_t bSet = PORTB;
	DDRC = 0X00;
	DDRD = 0x00;
	DDRB = 0x00;
	PORTB = 0X00;
	PORTD = 0X00;
	PORTC = 0X00;
    TCCR2B = 7; //this redundant register setting is used so that the corresponding ready flag in ASSR will let me know when the current TOSC cycle has completed.
    while (ASSR&0x01)
    {}
    
    EIMSK = 0x03;//enable external interrupt
    	sleep_enable();
		sleep_cpu();
    EIMSK = 0x03;//disable external interrupt
    //_delay_ms(100);
	DDRC = cDirection;
	DDRD = dDirection;
	DDRB = bDirection;
	PORTC = cSet;
	PORTD = dSet;
	PORTB = bSet;
}

int main(void)
{
    //Configure Ports
    DDRC = 0xFF; 
    DDRD = 0X13;
    DDRB = 0x30;
    

    PORTC = 0X08;
    PORTD = 0X01;
    PORTB = 0X20;
    
    //power-save sleep mode
    SMCR = 0x06; 
	

    
    
    //Display Refresher
    TCCR0A = 0x02; //CTC mode, interrupt on compare match
    TCCR0B = 0X04; //1/256 prescale
    TIMSK0 = 0x02; //enable timer0 compare match A
    OCR0A = 250;
    
    //Beeper
    TCCR1A = 0X00; //CTC mode, interrupt on compare match
    TCCR1B = 0X0B; //1/64 prescale
    TIMSK1 = 0X02; //enable timer0 compare match A
    OCR1A = 23;
    
    //timer2 keeps real time with the 32.768kHz crystal
    TCCR2B = 7; //clk/1024=>.125Hz interrupt with 32.768kHz crystal
    ASSR = 0x20; //enable asynchronous operation with external crystal
    TIMSK2 = 1; //enable overflow timer on timer 2
    
    //configure INT0 to active low
    EICRA=0x00;
    
    interfaceTimeout = maxInterfaceTimeout;
    printNum(threshold);
    sleepTime=maxSleepTime;
    sei();
    
    for(;;){
		
		uint8_t i = 0;
		//wait for the timer to time out or the user to push a button
		while((i<100)&&(PIND&buttonOneMask)&&(PIND&buttonTwoMask))
		{
			_delay_ms(1);
			i++;
		}
		
		
		//user is still holding the button
		if (!((PIND&buttonOneMask)&&(PIND&buttonTwoMask))) 
		{
			_delay_ms(debounceTime);
			
			//check buttons' status
			if (!(PIND&buttonOneMask))
			{
				if (interfaceTimeout>0) //don't change anything if you're just entering threshold change mode.
					incThreshold();
				interfaceTimeout=maxInterfaceTimeout;
				printNum(threshold);
				sleepTime=maxSleepTime;
				while (!(PIND&buttonOneMask))
				{} //wait for user to let go
			}
			else if (!(PIND&buttonTwoMask))
			{
				if (interfaceTimeout>0) //don't change anything if you're just entering threshold change mode.
					decThreshold();
				interfaceTimeout=maxInterfaceTimeout;
				printNum(threshold);
				sleepTime=maxSleepTime;
				while(!(PIND&buttonTwoMask))
				{} //wait for user to let go
			}
			
			
				
		}
		else
		{
			//if button is not pressed, take normal measurement and display it
			if (interfaceTimeout==0)
			{
				uint8_t temp = takeMeasurement(); 
				printNum(temp);
				
			}
			else
				interfaceTimeout--;
			//if it's time to sleep, sleep the unit
			if (sleepTime==0)
			{
				
				displayOn=0;
				_delay_ms(10);
				
				while(sleepTime==0)
				{
					measureTime=maxMeasureTime; //set the time between measurements
					while ((measureTime >0)&&(sleepTime==0))
					{
						goTheFuckToSleep();
						if (!(PIND&buttonOneMask)||!(PIND&buttonTwoMask))
						{
							printNum(threshold);
							displayOn=1;
							while(!(PIND&buttonOneMask)||!(PIND&buttonTwoMask))
							{} //wait for user to let go of wakeup button
						}
						
					}
						
				
					if (measureTime==0)
					{
						uint8_t temp=takeMeasurement();
						if (temp<threshold)
							{
								beepNow();
								_delay_ms(250);
							}
					}
				}
			}
				
				
					
				
			
		}
		
		
		
		
	}
    return 0;   /* never reached */
}

void printNum(uint8_t number)
{
    if (number>99)
        number=99;
    uint8_t tens = number/10;
    uint8_t ones = number%10;
    buffBTens = ((buffBTens&bMask)|(pgm_read_byte(&numbersB[tens])));
    buffBOnes = ((buffBOnes&bMask)|(pgm_read_byte(&numbersB[ones])));
    buffCTens = ((buffCTens&cMask)|(pgm_read_byte(&numbersC[tens])));
    buffCOnes = ((buffCOnes&cMask)|(pgm_read_byte(&numbersC[ones])));
    buffDTens = ((buffDTens&dMask)|(pgm_read_byte(&numbersD[tens])));
    buffDOnes = ((buffDOnes&dMask)|(pgm_read_byte(&numbersD[ones])));
}

ISR(TIMER0_COMPA_vect)
{
    if (displayOn)
    {
        if (PORTC&0X08)
        {
            PORTC = ((buffCTens & cdigitEnableMask)|0x04);
            PORTD = ((PORTD & dMask)| buffDTens);
            PORTB = ((PORTB & bMask)| buffBTens);
        }
        else
        {
            PORTC = ((buffCOnes & cdigitEnableMask)|0x08);
            PORTD = ((PORTD & dMask)|buffDOnes);
            PORTB = ((PORTB & bMask)|buffBOnes);
        }
    }
    else
    {
        PORTC = (PORTC | ~cdigitEnableMask);
    }
}



void beepNow()
{
    beepCount=700;
}
ISR(TIMER1_COMPA_vect)
{
    if (beepCount)
    {
        if (beepCount%2)
        {
            PORTD = PORTD & ~0X10;
        }
        else
        {
            PORTD = PORTD | 0x10;

        }
        

        beepCount=beepCount-1;

        
    }
    else
    {
        PORTD = PORTD & ~0X10; //Make sure to switch off current to the buzzer when you're done with the beep.
    }
}



ISR(INT0_vect)
{
	
	
	interfaceTimeout=maxInterfaceTimeout;
	sleepTime=maxSleepTime;
	EIMSK = 0x00;//disable external interrupt
	
	
}

ISR(INT1_vect)
{
	
	
	interfaceTimeout=maxInterfaceTimeout;
	sleepTime=maxSleepTime;
	EIMSK = 0x00;//disable external interrupt
	
	
}

ISR(TIMER2_OVF_vect)
{
    
      if (sleepTime>0)
		sleepTime--;
      if (measureTime>0)
		measureTime--;
      
	
}